using System;
using System.ServiceModel;
using System.ServiceModel.Dispatcher;
using System.Configuration;
using System.ServiceModel.Description;
using System.ServiceModel.Channels;
using System.ServiceModel.Security.Tokens;
using System.Collections.Generic;
using System.Security.Cryptography.X509Certificates;

[assembly: CLSCompliant(true)]
namespace Microsoft.ServiceModel.Samples.SecurityTokenService
{
    class Program
    {

        /// <summary>
        /// SAMPLE EXTENSIBILITY TASK:
        /// 
        /// Gets the signing certificate
        /// 
        /// Add your method to get the certificate
        /// 
        /// </summary>
        /// <returns></returns>
        public static X509Certificate2 SigningCertificate
        {
            get
            {
                string thumbprint = ConfigurationManager.AppSettings["certificateThumbprint"];
                if (string.IsNullOrEmpty(thumbprint))
                    throw new ArgumentException("thumbprint is not specified in configuration");

                X509Store store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
                store.Open(OpenFlags.ReadOnly);
                X509Certificate2Collection collection = store.Certificates.Find(X509FindType.FindByThumbprint, thumbprint, true);
                if (collection.Count != 1)
                {
                    throw new NotSupportedException("There must be exactly one certificate");
                }
                return collection[0];
            }
        }


        /// <summary>
        /// SAMPLE EXTENSIBILITY TASK:
        /// 
        /// Gets the issuer id
        /// 
        /// Add your code to get the issuer 
        /// 
        /// </summary>
        /// <returns></returns>
        public static string Issuer
        {
            get
            {
                string result = ConfigurationManager.AppSettings["issuer"];
                if (string.IsNullOrEmpty(result))
                    throw new ArgumentException("issuer is not specified in configuration");
                return result;
            }
        }


        static void Main(string[] args)
        {
            try
            {
                // Get base address from app settings in configuration
                Uri baseAddress = new Uri(ConfigurationManager.AppSettings["baseAddress"]);
                Uri baseMexAddress = new Uri(ConfigurationManager.AppSettings["baseMexAddress"]);

                List<ServiceHost> hostList = new List<ServiceHost>();
                try
                {
                    //
                    // Start the STSs
                    //
                    hostList.Add(StartSTS(typeof(CertificateAuthSTS), "smartcard", baseAddress, baseMexAddress));
                    hostList.Add(StartSTS(typeof(SelfIssuedSamlAuthSTS), "selfissuedsaml", baseAddress, baseMexAddress));
                    hostList.Add(StartSTS(typeof(UserNameAuthSTS), "usernamepassword", baseAddress, baseMexAddress));

                    Console.Write("\n\nPress <ENTER> to terminate services");
                    Console.ReadLine();
                }
                finally
                {
                    foreach (ServiceHost host in hostList)
                    {
                        host.Close();
                    }
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("An unexpected fault was encountered:\n{0}", e.ToString());
            }
        }

        static ServiceHost StartSTS(Type type, string stsLabel, Uri baseAddress, Uri baseMexAddress)
        {

            // Create the service host
            Uri stsAddress = new Uri(baseAddress.AbsoluteUri + "/" + stsLabel);
            ServiceHost serviceHost = new ServiceHost(type, stsAddress);
            
            // Don't require derived keys for the issue method
            ServiceEndpoint stsEndpoint = serviceHost.Description.Endpoints.Find(typeof(Microsoft.ServiceModel.Samples.SecurityTokenService.IWSTrustContract));
            BindingElementCollection bindingElements = stsEndpoint.Binding.CreateBindingElements();
            SecurityBindingElement sbe = bindingElements.Find<SecurityBindingElement>();
            RsaSecurityTokenParameters rsaParams = new RsaSecurityTokenParameters();
            rsaParams.InclusionMode = SecurityTokenInclusionMode.Never;
            rsaParams.RequireDerivedKeys = false;
            SupportingTokenParameters requirements = new SupportingTokenParameters();
            requirements.Endorsing.Add(rsaParams);
            sbe.OptionalOperationSupportingTokenParameters.Add(Microsoft.ServiceModel.Samples.SecurityTokenService.Constants.WSTrust.Actions.Issue, requirements);
            stsEndpoint.Binding = new CustomBinding(bindingElements);
            serviceHost.Credentials.ServiceCertificate.Certificate = SigningCertificate;
            
            // Add an https mex listener
            string mexAddress = baseMexAddress.AbsoluteUri + "/" + stsLabel + "/mex";
            serviceHost.AddServiceEndpoint(ServiceMetadataBehavior.MexContractName, MetadataExchangeBindings.CreateMexHttpsBinding(), mexAddress); 

            // Open the service
            serviceHost.Open();

            // Display the endpoints
            foreach (ChannelDispatcher cd in serviceHost.ChannelDispatchers)
            {
                foreach (EndpointDispatcher ed in cd.Endpoints)
                {
                    Console.WriteLine("Listener = {0}, State = {1}", ed.EndpointAddress.ToString(), cd.State.ToString());
                }
            }

            return serviceHost;
        }
    }
}