using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.Text;
using System.Threading;
using System.Xml;

namespace Client
{
    public class MessageInterceptor : IMessageInterceptor
    {
        private static Random randomizer = new Random(DateTime.Now.Millisecond);

        public void ProcessSend(ref Message message)
        {
            string action = message.Headers.Action;
            Console.WriteLine(string.Format("Voting on message {0} ...", action));
            if (!(action.Contains("ITradingService")))
            {
                return;
            }
            int randomNumber = MessageInterceptor.randomizer.Next(10);
            if (randomNumber > 6)
            {
                Console.WriteLine("Dropping message.");
                message = null;
                return;
            }
            Console.WriteLine("Forwarding message.");
        }

        public void ProcessReceive(ref Message message)
        {

        }
    }


    public interface IMessageInterceptor
    {
        void ProcessReceive(ref Message message);
        void ProcessSend(ref Message message);
    }

    public class InterceptorBindingElement : BindingElement
    {
        IMessageInterceptor interceptor = null;

        public InterceptorBindingElement()
        {
            this.interceptor = new MessageInterceptor();
        }

        protected InterceptorBindingElement(InterceptorBindingElement other)
            : base(other)
        {
            this.interceptor = other.interceptor;
        }

        public override T GetProperty<T>(BindingContext context)
        {
            return context.GetInnerProperty<T>();
        }
        public override IChannelFactory<TChannel> BuildChannelFactory<TChannel>(BindingContext context)
        {
            InterceptorChannelFactory<TChannel> result = new InterceptorChannelFactory<TChannel>(this.interceptor, context.Binding);
            
            result.InnerChannelFactory = context.BuildInnerChannelFactory<TChannel>();
            return result;
        }

        public override bool CanBuildChannelFactory<TChannel>(BindingContext context)
        {
            return base.CanBuildChannelFactory<TChannel>(context);
        }

        public override BindingElement Clone()
        {
            return new InterceptorBindingElement(this);
        }
    }



    class NullMessageInterceptor : IMessageInterceptor
    {
        public void ProcessSend(ref Message message)
        {
        }

        public void ProcessReceive(ref Message message)
        {
        }
    }

    public class InterceptorChannelFactory<TChannel> : ChannelFactoryBase<TChannel>
    {
        IMessageInterceptor interceptor;

        public InterceptorChannelFactory(IMessageInterceptor interceptor, IDefaultCommunicationTimeouts timeouts)
            : base(timeouts)
        {
            if (interceptor == null)
            {
                this.interceptor = new NullMessageInterceptor();
            }
            else
            {
                this.interceptor = interceptor;
            }

        }

        public IMessageInterceptor Interceptor
        {
            get
            {
                return interceptor;
            }
        }

        protected override IAsyncResult OnBeginOpen(TimeSpan timeout, AsyncCallback callback, object state)
        {

            return new OpenAsyncResult(callback, state);
        }

        protected override void OnEndOpen(IAsyncResult result)
        {
            
        }
       
        
        protected override Type GetCommunicationObjectType()
        {
            return typeof(TChannel);
        }

       
        protected override TChannel OnCreateChannel(EndpointAddress remoteAddress, Uri via)
        {
            InnerChannelFactory.Open();
            TChannel innerChannel = InnerChannelFactory.CreateChannel(remoteAddress, via);
            if (typeof(TChannel) == typeof(IOutputChannel))
            {
                return (TChannel)(object)new InterceptorOutputChannel(this, (IOutputChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IRequestChannel))
            {
                return (TChannel)(object)new InterceptorRequestChannel(this, (IRequestChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IOutputSessionChannel))
            {
                return (TChannel)(object)new InterceptorOutputSessionChannel(this, (IOutputSessionChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IRequestSessionChannel))
            {
                return (TChannel)(object)new InterceptorRequestSessionChannel(this,
                    (IRequestSessionChannel)innerChannel);
            }
            throw new ArgumentException("Channel type is not supported.", "TChannel");
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            ThrowIfInnerFactoryNotSet();
        }

        void ThrowIfInnerFactoryNotSet()
        {
            if (this.InnerChannelFactory == null)
            {
                throw new InvalidOperationException("Inner channel factory is not set.");
            }
        }
        IChannelFactory<TChannel> innerChannelFactory;

        public IChannelFactory<TChannel> InnerChannelFactory
        {
            get { return innerChannelFactory; }
            set { innerChannelFactory = value; }
        }
   

        class InterceptorOutputChannel : InterceptorChannelBase<IOutputChannel>, IOutputChannel
        {
            InterceptorChannelFactory<TChannel> factory;

            public InterceptorOutputChannel(InterceptorChannelFactory<TChannel> factory, IOutputChannel innerChannel)
                : base(factory, innerChannel, factory.interceptor)
            {
                this.factory = factory;
            }

            public EndpointAddress RemoteAddress
            {
                get { return this.InnerChannel.RemoteAddress; }
            }

            public Uri Via
            {
                get { return this.InnerChannel.Via; }
            }

            public IAsyncResult BeginSend(Message message, AsyncCallback callback, object state)
            {
                return this.BeginSend(message, factory.DefaultSendTimeout, callback, state);
            }

            public IAsyncResult BeginSend(Message message, TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new SendAsyncResult<IOutputChannel>(this, message, timeout, callback, state);
            }

            public void EndSend(IAsyncResult result)
            {
                SendAsyncResult<IOutputChannel>.End(result);
            }

            public void Send(Message message)
            {
                this.Send(message, factory.DefaultSendTimeout);
            }

            public void Send(Message message, TimeSpan timeout)
            {
                ThrowIfDisposedOrNotOpen();
                Interceptor.ProcessSend(ref message);
                if (message == null)
                {
                    OnDropMessage();
                }
                else
                {
                    this.InnerChannel.Send(message, timeout);
                }
            }
        }

        class InterceptorRequestChannel : InterceptorChannelBase<IRequestChannel>, IRequestChannel
        {
            InterceptorChannelFactory<TChannel> factory;

            public InterceptorRequestChannel(InterceptorChannelFactory<TChannel> factory, IRequestChannel innerChannel)
                : base(factory, innerChannel, factory.interceptor)
            {
                this.factory = factory;
            }

            public EndpointAddress RemoteAddress
            {
                get { return this.InnerChannel.RemoteAddress; }
            }

            public Uri Via
            {
                get { return this.InnerChannel.Via; }
            }

            public IAsyncResult BeginRequest(Message message, AsyncCallback callback, object state)
            {
                return this.BeginRequest(message, factory.DefaultSendTimeout, callback, state);
            }

            public IAsyncResult BeginRequest(Message message, TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new RequestAsyncResult(this, message, timeout, callback, state);
            }

            public Message EndRequest(IAsyncResult result)
            {
                return RequestAsyncResult.End(result);
            }

            public Message Request(Message message)
            {
                return this.Request(message, factory.DefaultSendTimeout);
            }

            public Message Request(Message message, TimeSpan timeout)
            {
                ThrowIfDisposedOrNotOpen();
                Interceptor.ProcessSend(ref message);
                if (message == null)
                {
                    return null;
                }
                else
                {
                    Message reply = this.InnerChannel.Request(message, timeout);
                    Interceptor.ProcessReceive(ref reply);

                    if (reply == null)
                    {
                        OnDropMessage();
                        return Message.CreateMessage(MessageVersion.Soap11WSAddressing10, new DroppedMessageFault("Reply Message dropped by the interceptor."), null);
                    }
                    else
                    {
                        return reply;
                    }
                }
            }
        }



        class InterceptorOutputSessionChannel : InterceptorOutputChannel, IOutputSessionChannel
        {
            IOutputSessionChannel innerSessionChannel;

            public InterceptorOutputSessionChannel(InterceptorChannelFactory<TChannel> factory, IOutputSessionChannel innerChannel)
                : base(factory, innerChannel)
            {
                this.innerSessionChannel = innerChannel;
            }

            public IOutputSession Session
            {
                get { return innerSessionChannel.Session; }
            }

            internal override void OnDropMessage()
            {
                Fault();
                innerSessionChannel.Abort();
            }
        }

        class InterceptorRequestSessionChannel : InterceptorRequestChannel, IRequestSessionChannel
        {
            IRequestSessionChannel innerSessionChannel;

            public InterceptorRequestSessionChannel(InterceptorChannelFactory<TChannel> factory,
                IRequestSessionChannel innerChannel)
                : base(factory, innerChannel)
            {
                this.innerSessionChannel = innerChannel;
            }

            public IOutputSession Session
            {
                get { return innerSessionChannel.Session; }
            }

            internal override void OnDropMessage()
            {
                Fault();
                innerSessionChannel.Abort();
            }
        }


    }

    public class InterceptorChannelBase<TChannel> : ChannelBase where TChannel : IChannel
    {
        TChannel innerChannel;
        IMessageInterceptor interceptor;

        protected InterceptorChannelBase(ChannelManagerBase manager, TChannel innerChannel, IMessageInterceptor interceptor)
            : base(manager)
        {
            this.innerChannel = innerChannel;
            this.interceptor = interceptor;
        }

        internal TChannel InnerChannel
        {
            get { return this.innerChannel; }
        }

        internal IMessageInterceptor Interceptor
        {
            get { return this.interceptor; }
        }

        protected override void OnAbort()
        {
            this.innerChannel.Abort();
        }

        protected override IAsyncResult OnBeginClose(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginClose(timeout, callback, state);
        }

        protected override IAsyncResult OnBeginOpen(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginOpen(timeout, callback, state);
        }

        protected override void OnClose(TimeSpan timeout)
        {
            this.innerChannel.Close(timeout);
        }

        protected override void OnEndClose(IAsyncResult result)
        {
            this.innerChannel.EndClose(result);
        }

        protected override void OnEndOpen(IAsyncResult result)
        {
            this.innerChannel.EndOpen(result);
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            this.innerChannel.Open(timeout);
        }

        internal virtual void OnDropMessage() { }
    }

    public class SendAsyncResult<TChannel> : AsyncResult where TChannel : IOutputChannel
    {
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<TChannel> channel;

        public SendAsyncResult(InterceptorChannelBase<TChannel> channel, Message message, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                Complete(true);
            }
            else
            {
                channel.InnerChannel.BeginSend(this.message, timeout, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                channel.InnerChannel.EndSend(asyncResult);
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static void End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            SendAsyncResult<TChannel> outputResult = result as SendAsyncResult<TChannel>;
            if (outputResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(outputResult);
        }
    }

    public abstract class AsyncResult : IAsyncResult
    {
        AsyncCallback callback;
        object state;
        bool completedSynchronously;
        bool endCalled;
        Exception exception;
        bool isCompleted;
        ManualResetEvent manualResetEvent;

        protected AsyncResult(AsyncCallback callback, object state)
        {
            this.callback = callback;
            this.state = state;
        }

        public object AsyncState
        {
            get { return state; }
        }

        public WaitHandle AsyncWaitHandle
        {
            get
            {
                if (manualResetEvent != null)
                {
                    return manualResetEvent;
                }

                lock (ThisLock)
                {
                    if (manualResetEvent == null)
                    {
                        manualResetEvent = new ManualResetEvent(isCompleted);
                    }
                }

                return manualResetEvent;
            }
        }

        public bool CompletedSynchronously
        {
            get { return completedSynchronously; }
        }

        public bool IsCompleted
        {
            get { return isCompleted; }
        }

        object ThisLock
        {
            get { return this; }
        }

        protected void Complete(bool completedSynchronously)
        {
            if (isCompleted)
            {
                throw new InvalidOperationException("AsyncResults can only be completed once.");
            }
            ManualResetEvent manualResetEvent = null;
            this.completedSynchronously = completedSynchronously;

            if (completedSynchronously)
            {
                // If we completedSynchronously, then there's no chance that the manualResetEvent was created so
                // we don't need to worry about a race
                this.isCompleted = true;
                if (this.manualResetEvent != null)
                {
                    throw new InvalidOperationException("No ManualResetEvent should be created for a synchronous AsyncResult.");
                }
            }
            else
            {
                lock (ThisLock)
                {
                    this.isCompleted = true;
                    manualResetEvent = this.manualResetEvent;
                }
            }

            try
            {
                if (callback != null)
                {
                    callback(this);
                }
                if (manualResetEvent != null)
                {
                    manualResetEvent.Set();
                }
            }
            catch (Exception unhandledException)
            {
                // The callback raising an exception is equivalent to Main raising an exception w/out a catch.
                // Queue it onto another thread and throw it there to ensure that there's no other handler in 
                // place and the default unhandled exception behavior occurs.

                // Because the stack trace gets lost on a rethrow, we're wrapping it in a generic exception
                // so the stack trace is preserved.
                unhandledException = new Exception("AsyncCallbackException", unhandledException);
                ThreadPool.UnsafeQueueUserWorkItem(new WaitCallback(RaiseUnhandledException), unhandledException);
            }
        }

        protected void Complete(bool completedSynchronously, Exception exception)
        {
            this.exception = exception;
            Complete(completedSynchronously);
        }

        protected static void End(AsyncResult asyncResult)
        {
            if (asyncResult == null)
            {
                throw new ArgumentNullException("asyncResult");
            }

            if (asyncResult.endCalled)
            {
                throw new InvalidOperationException("Async object already ended.");
            }

            asyncResult.endCalled = true;

            if (!asyncResult.isCompleted)
            {
                using (WaitHandle waitHandle = asyncResult.AsyncWaitHandle)
                {
                    waitHandle.WaitOne();
                }
            }

            if (asyncResult.exception != null)
            {
                throw asyncResult.exception;
            }
        }

        void RaiseUnhandledException(object o)
        {
            Exception exception = (Exception)o;
            throw exception;
        }
    }

    public class OpenAsyncResult : AsyncResult
    {
        public OpenAsyncResult(AsyncCallback callback, object state) : base(callback, state)
        {}
    }
    public class RequestAsyncResult : AsyncResult
    {
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<IRequestChannel> channel;

        public RequestAsyncResult(InterceptorChannelBase<IRequestChannel> channel, Message message, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                Complete(true);
            }
            else
            {
                channel.InnerChannel.BeginRequest(this.message, timeout, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                Message reply = channel.InnerChannel.EndRequest(asyncResult);
                channel.Interceptor.ProcessReceive(ref reply);
                if (reply == null)
                {
                    channel.OnDropMessage();
                    this.message = Message.CreateMessage(MessageVersion.Soap11WSAddressing10,new DroppedMessageFault("Reply Message dropped by interceptor."),null);
                }
                else
                {
                    this.message = reply;
                }
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static Message End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            RequestAsyncResult requestResult = result as RequestAsyncResult;
            if (requestResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(requestResult);

            return requestResult.message;
        }
    }

    public class ReplyAsyncResult : AsyncResult
    {
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<IReplyChannel> channel;

        public ReplyAsyncResult(InterceptorChannelBase<IReplyChannel> channel, Message message,
            TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            //this.innerContext = innerContext;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                this.message = Message.CreateMessage(MessageVersion.Soap11WSAddressing10,new DroppedMessageFault("Reply Message dropped by interceptor."), null);
                Complete(true);
            }
            else
            {
                
               channel.InnerChannel.ReceiveRequest().BeginReply(this.message, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                channel.InnerChannel.ReceiveRequest().EndReply(asyncResult);
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static void End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            ReplyAsyncResult requestResult = result as ReplyAsyncResult;
            
            if (requestResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(requestResult);
        }
    }

    public class DroppedMessageFault : MessageFault
    {
        FaultReason faultReason;
        FaultCode faultCode;

        public DroppedMessageFault(string reason)
        {
            faultReason = new FaultReason(reason);
            faultCode = new FaultCode("DroppedMessageFault");
        }

        public override FaultCode Code
        {
            get { return faultCode; }
        }

        public override bool HasDetail
        {
            get { return false; }
        }

        public override FaultReason Reason
        {
            get { return faultReason; }
        }
        protected override void OnWriteDetail(XmlDictionaryWriter writer, EnvelopeVersion version)
        {
            throw new NotImplementedException("The method or operation is not implemented.");
        }

        protected override void OnWriteDetailContents(XmlDictionaryWriter writer)
        {
            throw new Exception("The method or operation is not implemented.");
        }
 
    }
}
